/* Copyright (c) 2005, Nigel Winterbottom
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.
   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. */

/*
  eeprom1.S

  Contributors:
	Created by Nigel Winterbttom <nigel@n-winterbottom.freeserve.co.uk>
	Using original work by Marek Michalkiewicz <marekm@linux.org.pl>

	This optimised eeprom library reduces eeprom cell wear and increases speed by
	suppression of a write operation when the location already contains the data
	to be written.

	It is compatible with the header file (eeprom.h) written by Marek.
*/

#include "macros.inc"
#include "ctoasm.inc"

/* the same library is used for 2313 and 8515 for now -
   I hope writing 0 to non-existent EEARH doesn't hurt... */

;	#ifndef EEARH
;	#define EEARH (EEARL+1)
;	#endif


	.section .text

#define wrval rP1
#define rdval rP0

#ifdef L_eeRdWr

; Internal routine to read one byte from EEPROM. Address in rP1:rP0 or Z depending on entry point

	.global Rd
Rd:
1:	sbic	_SFR_IO_ADDR(EECR), EEWE
	rjmp	1b						; Make sure EEPROM is ready

	in		__tmp_reg__, _SFR_IO_ADDR(SREG)
	cli										; No ints during eeprom read
#ifdef EEARH
	out		_SFR_IO_ADDR(EEARH), r31
#endif
	out		_SFR_IO_ADDR(EEARL), r30
	sbi		_SFR_IO_ADDR(EECR), EERE
	in		rdval, _SFR_IO_ADDR(EEDR)
	out		_SFR_IO_ADDR(SREG), __tmp_reg__
	ret

; Internal routine to write one byte to EEPROM. Address in Z, data in (wrval)
; Write is not performed if not necessary

	.global Wr
Wr:
	rcall	Rd								; Read first
	cp		rdval, wrval					; IF rdval==wrval THEN skip write
	breq	1f
	in		__tmp_reg__, _SFR_IO_ADDR(SREG)
	cli										; No ints between setting EEMWE and EEWE
	out		_SFR_IO_ADDR(EEDR), wrval
	sbi		_SFR_IO_ADDR(EECR), EEMWE
	sbi		_SFR_IO_ADDR(EECR), EEWE
	out		_SFR_IO_ADDR(SREG), __tmp_reg__
1:	ret
#endif /* L_eeRdWr */


#ifdef L_eeprom_read_byte
/* read one byte from EEPROM */
/* uint8_t eeprom_read_byte(const uint8_t *addr); */
/* addr = r25:r24, result = r25(=0):r24 */

	.global _U(eeprom_read_byte)

_U(eeprom_read_byte):
	LOAD_Z(rP1,rP0)
	rcall	Rd
	mov		rP1, rdval
	clr		rP0
	ret
#endif /* L_eeprom_read_byte */



#ifdef L_eeprom_read_word
/* read a little endian 16-bit word from EEPROM */
/* uint16_t eeprom_read_word(const uint16_t *addr); */
/* addr = r25:r24, result = r25:r24 */

	.global _U(eeprom_read_word)

_U(eeprom_read_word):
	LOAD_Z(rP1,rP0)
	rcall	Rd
	mov		rP1, rdval
	adiw	r30,1					; Inc Address
;	subi	r30, lo8(-1)			; Inc Address
;	sbci	r31, hi8(-1)
	rjmp	Rd
#endif /* L_eeprom_read_word */



#ifdef L_eeprom_write_byte
/* write a byte to EEPROM */
/* void eeprom_write_byte(uint8_t *addr, uint8_t val); */
/* addr = r25:r24, val = r22 */
#define val rP3

	.global _U(eeprom_write_byte)

_U(eeprom_write_byte):
	LOAD_Z(rP1,rP0)
	mov		wrval, val
	rjmp	Wr

#undef val
#endif /* L_eeprom_write_byte */



#ifdef L_eeprom_write_word
/* write a word to EEPROM */
/* void eeprom_write_word (uint16_t *addr, uint16_t val); */
/* addr = r25:r24, val = r23:r22 */

#define val_hi rP2
#define val_lo rP3

	.global _U(eeprom_write_word)

_U(eeprom_write_word):
	rcall	_U(eeprom_write_byte)	; Note: val already setup as val_lo
	adiw	r30,1					; Inc Address
;	subi	r30, lo8(-1)			; Inc Address
;	sbci	r31, hi8(-1)
	mov		wrval, val_hi
	rjmp	Wr

#undef val_hi
#undef val_lo
#endif /* L_eeprom_write_word */




#ifdef L_eeprom_read_block
/* read a block of bytes from EEPROM */
/* void eeprom_read_block(void *buf, const void *addr, size_t n); */
/* buf = r25:r24, addr = r23:r22, n = r21:r20 */

#define buf_hi rP0
#define buf_lo rP1
#define addr_hi rP2
#define addr_lo rP3
#define n_hi rP4
#define n_lo rP5

	.global _U(eeprom_read_block)

_U(eeprom_read_block):
	cp		n_lo, __zero_reg__
	cpc		n_hi, __zero_reg__
	breq	2f
	LOAD_X(buf_lo, buf_hi)
	LOAD_Z(addr_lo, addr_hi)

1:	rcall	Rd
	adiw	r30,1					; Inc Address
;	subi	r30, lo8(-1)			; Inc Address
;	sbci	r31, hi8(-1)
	st		X+, rdval
	subi	n_lo, lo8(1)
	sbci	n_hi, hi8(1)
	brne	1b
2:	ret

#undef buf_hi
#undef buf_lo
#undef addr_hi
#undef addr_lo
#undef n_hi
#undef n_lo
#endif /* L_eeprom_read_block */



#ifdef L_eeprom_write_block
/* write a block of bytes to EEPROM */
/* void eeprom_write_block (const void *buf, void *addr, size_t n); */
/* buf = r25:r24, addr = r23:r22, n = r21:r20 */

#define buf_hi rP0
#define buf_lo rP1
#define addr_hi rP2
#define addr_lo rP3
#define n_hi rP4
#define n_lo rP5

	.global _U(eeprom_write_block)

_U(eeprom_write_block):
	cp		n_lo, __zero_reg__			; Check if really there is something to write
	cpc		n_hi, __zero_reg__
	breq	2f
	LOAD_X(buf_lo, buf_hi)
	LOAD_Z(addr_lo, addr_hi)

1:	ld		wrval, X+
	rcall	Wr
	adiw	r30,1					; Inc Address
;	subi	r30, lo8(-1)			; Inc Address
;	sbci	r31, hi8(-1)
	subi	n_lo, lo8(1)
	sbci	n_hi, hi8(1)
	brne	1b
2:	ret

#undef buf_hi
#undef buf_lo
#undef addr_hi
#undef addr_lo
#undef n_hi
#undef n_lo
#endif /* L_eeprom_write_block */

.end
